/*
 * Copyright 2021 Nilanjan Daw, Synerg Lab, Department of CSE, IIT Bombay
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * This Java source file was generated by the Gradle 'init' task.
 */
package HpdosClient;

import HpdosClient.MessageFormat.MessageConstants;
import HpdosClient.lib.StorageModel;
import HpdosClient.lib.StorageService;
import com.google.common.base.Stopwatch;
import com.google.common.util.concurrent.AtomicDouble;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.MoreExecutors;
import hpdos.grpc.Packet;
import hpdos.grpc.Response;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.concurrent.*;

public class ClientRunner {
    public static int runtime;
    private final String clientID;
    public static String propertiesFile;
    private int cCreate, cRead, cUpdate, cDelete;
    public boolean experimentEnded = false;
    private Queue<Long> createTime, updateTime, readTime, deleteTime;
    private final Properties properties;
    private final List<StorageModel> generatedPacket;
    private int threadCount;
    private int rps;
    AtomicDouble openCount;
    private boolean cacheEnabled;

    public ClientRunner() {
        clientID = UUID.randomUUID().toString();
        generatedPacket = Collections.synchronizedList(new ArrayList<>());
        properties = new Properties();
        threadCount = Runtime.getRuntime().availableProcessors();
        try {
            openCount = new AtomicDouble(0);
            InputStream inputStream = new FileInputStream(propertiesFile);
            this.properties.load(inputStream);
            System.out.println((String) properties.get("app.concurrency"));
            this.rps = Integer.parseInt((String) properties.get("app.rps"));
            this.threadCount = Integer.parseInt((String) properties.get("app.thread_count"));
            runtime = Integer.parseInt((String) properties.get("app.runtime"));
            this.cacheEnabled = properties.get("app.cache").equals("1");

            this.cCreate = Integer.parseInt((String) properties.get("app.cycle_create"));
            this.cRead = Integer.parseInt((String) properties.get("app.cycle_read"));
            this.cUpdate = Integer.parseInt((String) properties.get("app.cycle_update"));
            this.cDelete = Integer.parseInt((String) properties.get("app.cycle_delete"));

            this.createTime = new ConcurrentLinkedQueue<>();
            this.updateTime = new ConcurrentLinkedQueue<>();
            this.readTime = new ConcurrentLinkedQueue<>();
            this.deleteTime = new ConcurrentLinkedQueue<>();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    public String getGreeting() {
        return "Hello World!";
    }

    private String createString() {
        double dataSize = Double.parseDouble((String) properties.get("app.data_size"));
        dataSize /= 2.0; // Java strings are 2B long
        String conversionFactor = (String) properties.get("app.data_conversion_factor");
        int multiplier = 1;
        switch (conversionFactor) {
            case "G": multiplier *= 1000;
            case "M": multiplier *= 1000;
            case "K": multiplier *= 1000;
        }
        char[] data = new char[(int)(dataSize * multiplier)];
        return new String(data);
    }

    private void seedServer(StorageService storageService, String value) throws InterruptedException {
        CountDownLatch latch = new CountDownLatch(10);
        for (int i = 0; i < 10; i++) {
            String key = Integer.toString((int) (Math.random() * Integer.MAX_VALUE));
            ListenableFuture<Packet> res = storageService.create(key, value, MessageConstants.METADATA_ACCESS_PRIVATE);
            StorageModel model = new StorageModel(0, value.length(), key,
                    MessageConstants.METADATA_ACCESS_PRIVATE, clientID, 0, value);
            res.addListener(() -> {
                try {
                    Packet packet = res.get();
                    for (Response response: packet.getResponseList()) {
                        if (response.getStatus() == MessageConstants.STATUS_OK) {
                            model.updateData(response.getAck());
                            generatedPacket.add(model);
                        }
                    }
                } catch (InterruptedException | ExecutionException e) {
                    e.printStackTrace();
                }
                latch.countDown();
            }, MoreExecutors.directExecutor());
        }
        latch.await();
    }

    private void seedServer(StorageService storageService, String value, String threadId) throws InterruptedException {
        CountDownLatch latch = new CountDownLatch(10);
        for (int i = 0; i < 10; i++) {
            String key = threadId + (int) (Math.random() * Integer.MAX_VALUE);
            ListenableFuture<Packet> res = storageService.create(key, value, MessageConstants.METADATA_ACCESS_PRIVATE);
            StorageModel model = new StorageModel(0, value.length(), key,
                    MessageConstants.METADATA_ACCESS_PRIVATE, clientID, 0, value);
            res.addListener(() -> {
                try {
                    Packet packet = res.get();
                    for (Response response: packet.getResponseList()) {
                        if (response.getStatus() == MessageConstants.STATUS_OK) {
                            model.updateData(response.getAck());
                            generatedPacket.add(model);
                        }
                    }
                } catch (InterruptedException | ExecutionException e) {
                    e.printStackTrace();
                }
                latch.countDown();
            }, MoreExecutors.directExecutor());
        }
        latch.await();
    }

    public double runExperimentByRps(String threadId) throws InterruptedException, ExecutionException {
        String value = createString(), updatedValue = createString();
        StorageService storageService = new StorageService(clientID, this.cacheEnabled);
        storageService.initStorage();
        this.seedServer(storageService, value, threadId);

        double totalBracket = this.cCreate + this.cRead + this.cUpdate + this.cDelete;
        double createBracket, readBracket, updateBracket;

        createBracket = this.cCreate * 1.0 / totalBracket;
        readBracket = createBracket + this.cRead / totalBracket;
        updateBracket = readBracket + this.cUpdate / totalBracket;
        long sleepTime = (threadCount * 1000L) / this.rps;
        Thread.sleep((long)(Math.random() * 500)); // randomizing start time to avoid spikes in request
        do {
            openCount.addAndGet(1);
            double toss = Math.random();
            if (toss < createBracket) {

                String key = threadId + (int) (Math.random() * Integer.MAX_VALUE);
                StorageModel model = new StorageModel(0, value.length(), key,
                        MessageConstants.METADATA_ACCESS_PRIVATE, clientID, 0, value);
                final long createStart = System.currentTimeMillis();
                ListenableFuture<Packet> res = storageService.create(model.getKey(), model.getValue(), model.getAccessType());
                res.addListener(() -> {
                    try {
                        Packet packet = res.get();
                        this.createTime.add(System.currentTimeMillis() - createStart);
                        for (Response response: packet.getResponseList()) {
                            if (response.getStatus() == MessageConstants.STATUS_OK) {
                                model.updateData(response.getAck());
                                generatedPacket.add(model);
                            }

                        }
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }

                    openCount.addAndGet(-1);
                }, MoreExecutors.directExecutor());
            } else if (toss < readBracket) {
                int index; StorageModel sendPacket;
                synchronized (generatedPacket) {
                    index = (int) (Math.random() * generatedPacket.size());
                    sendPacket = this.generatedPacket.get(index);
                }
                final long readStart = System.currentTimeMillis();

                try {
                    ListenableFuture<StorageModel> res = storageService.read(sendPacket.getKey());
                    res.addListener(() -> {
                        try {
                            StorageModel model = res.get();
                            readTime.add(System.currentTimeMillis() - readStart);
                            synchronized (sendPacket) {
                                sendPacket.updateData(model);
                            }
                        } catch (Exception e) {

                        }

                        openCount.addAndGet(-1);
                    }, MoreExecutors.directExecutor());
                } catch (IOException e) {
                    e.printStackTrace();
                }

            } else if (toss < updateBracket) {
                int index; StorageModel sendPacket;
                synchronized (generatedPacket) {
                    index = (int) (Math.random() * generatedPacket.size());
                    sendPacket = this.generatedPacket.get(index);
                }
                final long updateStart = System.currentTimeMillis();
                ListenableFuture<Packet> res = storageService.update(sendPacket.getKey(), updatedValue,
                        sendPacket.getVersion());
                res.addListener(() -> {
                    try {

                        Packet packet = res.get();
                        updateTime.add(System.currentTimeMillis() - updateStart);
//                        System.out.println("packet " + packet);
                        for (Response response: packet.getResponseList()) {
                            if (response.getStatus() == MessageConstants.STATUS_OK)
                                if (sendPacket.getKey().equals(response.getAck().getKey())) {
                                    synchronized (sendPacket) {
                                        sendPacket.updateData(response.getAck());
                                    }
                                }
                        }

                        openCount.addAndGet(-1);
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }
                }, MoreExecutors.directExecutor());
            } else {
//                System.out.println("delete");
                int index; StorageModel sendPacket;
                synchronized (generatedPacket) {
                    index = (int) (Math.random() * generatedPacket.size());
                    sendPacket = this.generatedPacket.get(index);
                }
                final long deleteStart = System.currentTimeMillis();
                ListenableFuture<Packet> res = storageService.delete(sendPacket.getKey(), sendPacket.getVersion());
                res.addListener(() -> {
                    try {
                        Packet packet = res.get();
                        deleteTime.add(System.currentTimeMillis() - deleteStart);
//                        System.out.println("packet " + packet);
                        for (Response response: packet.getResponseList()) {
                            if (response.getStatus() == MessageConstants.STATUS_OK)
                                if (sendPacket.getKey().equals(response.getAck().getKey())) {
                                    synchronized (generatedPacket) {
                                        generatedPacket.remove(sendPacket);
                                    }
                                }
                        }
                        openCount.addAndGet(-1);
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }
                }, MoreExecutors.directExecutor());
            }
            while (openCount.get() > this.rps)
                Thread.sleep(sleepTime);
            Thread.sleep(sleepTime);
        } while (!this.experimentEnded);
        storageService.cleanup();
        return 0;
    }


    private void timerService() {
        Stopwatch stopwatch = Stopwatch.createUnstarted();
        stopwatch.start();
        Timer timer = new Timer();
        timer.scheduleAtFixedRate(new TimerTask() {
            @Override
            public void run() {
                if (experimentEnded) {
                    timer.cancel();
                    timer.purge();
                }
                System.out.println("Experiment ran: " + stopwatch + " openCount: " + openCount);
            }
        }, 5000, 5000);
    }

    private void printStatistics(double totalRuntime) {
        long readQps = 0, createQps = 0, updateQps = 0, deleteQps = 0;
        double avgRead = 0, avgCreate = 0, avgUpdate = 0, avgDelete = 0;
        for (Long time: this.readTime) {
            readQps++;
            avgRead += time;
        }
        avgRead /= readQps * 1.0;
        for (Long time: this.createTime) {
            createQps++;
            avgCreate += time;
        }
        avgCreate /= createQps * 1.0;
        for (Long time: this.updateTime) {
            updateQps++;
            avgUpdate += time;
        }
        avgUpdate /= updateQps * 1.0;
        for (Long time: this.deleteTime) {
            deleteQps++;
            avgDelete += time;
        }
        avgDelete /= deleteQps * 1.0;
        double totalQps = readQps + createQps + updateQps + deleteQps;
        System.out.println("Total runtime: " + totalRuntime);
        System.out.println("Read: " + readQps + " Create: " + createQps
                + " Update: " + updateQps + " Delete: " + deleteQps + " Total: " + totalQps);
        totalRuntime /= 1000;
        System.out.println("Total QPS: " + totalQps / totalRuntime + " avg query time: " +
                (totalQps * threadCount / (totalRuntime)));
        System.out.println("Read QPS: " + readQps / totalRuntime + " avg query time: " + avgRead);
        System.out.println("Create QPS: " + createQps / totalRuntime + " avg query time: " + avgCreate);
        System.out.println("Update QPS: " + updateQps / totalRuntime + " avg query time: " + avgUpdate);
        System.out.println("Delete QPS: " + deleteQps / totalRuntime + " avg query time: " + avgDelete);
    }

    private void cleanupExperiment(StorageService storageService) throws InterruptedException {
        System.out.println("Cleaning up remnants");
        CountDownLatch latch = new CountDownLatch(generatedPacket.size());
        for (StorageModel data: generatedPacket) {
            ListenableFuture<Packet> res = storageService.delete(data.getKey(), data.getVersion());
            res.addListener(latch::countDown, MoreExecutors.directExecutor());
        }
        latch.await();
        storageService.cleanup();
    }
    public static void main(String[] args) throws InterruptedException, ExecutionException {
        propertiesFile = args[0];
        ClientRunner clientRunner = new ClientRunner();
        System.out.println(clientRunner.getGreeting());
        StorageService storageService = new StorageService(clientRunner.clientID, clientRunner.cacheEnabled);
        storageService.initStorage();
        System.out.println("storage initialised");
        clientRunner.seedServer(storageService, clientRunner.createString());
        System.out.println("server seeded. Length: " + clientRunner.generatedPacket.size());
        final long startTime = System.currentTimeMillis();
        clientRunner.timerService();
        ExecutorService executorService = Executors.newSingleThreadExecutor();
        executorService.submit(() -> {
            try {
                Thread.sleep(runtime * 1000L);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            clientRunner.experimentEnded = true;
            System.out.println("Experiment ended");
        });
        ExecutorService experimentExecutors = Executors.newFixedThreadPool(clientRunner.threadCount);
        Set<Callable<Double>> callables = new HashSet<>();
        System.out.println("Starting experiment");
        for (int i = 1; i <= clientRunner.threadCount; i++) {
            int finalI = i;
            callables.add(() -> clientRunner.runExperimentByRps(Integer.toString(finalI)));

        }
        List<Future<Double>> futures = experimentExecutors.invokeAll(callables);
        for (Future<Double> future: futures) {
            future.get();
        }
        long endTime = System.currentTimeMillis();
        System.out.println("Waiting for system to settle down");
        double totalRuntime = endTime - startTime;
        clientRunner.printStatistics(totalRuntime);
        clientRunner.cleanupExperiment(storageService);
        System.out.println(storageService.getCache().stats());
        executorService.shutdown();
        boolean status = executorService.awaitTermination(1000, TimeUnit.MICROSECONDS);
        if (!status)
            executorService.shutdownNow();
        experimentExecutors.shutdownNow();
    }
}
