/*
 * Copyright 2021 Nilanjan Daw, Synerg Lab, Department of CSE, IIT Bombay
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package HpdosClient.lib;

import HpdosClient.ConfigConstants;
import HpdosClient.MessageFormat.MessageConstants;
import HpdosClient.MessageFormat.RequestBuilder;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.util.concurrent.ListenableFuture;
import hpdos.grpc.*;
import io.grpc.ManagedChannel;
import io.grpc.ManagedChannelBuilder;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.*;

public class StorageService {

    private final String clientID;
    private ManagedChannel masterChannel;
    private ArrayList<ManagedChannel> channels;

    private List<Follower> replicaSet;
    private LoadingCache<String, StorageModel> cache = null;
    private final boolean cacheEnabled;
    public StorageService(String clientID, boolean cacheEnabled) {
        this.clientID = clientID;
        this.cacheEnabled = cacheEnabled;
        if (this.cacheEnabled)
            this.cache = CacheBuilder
                    .newBuilder()
                    .maximumSize(10000)
                    .recordStats()
                    .build(new CacheLoader<>() {
                        @Override
                        public StorageModel load(@Nonnull String key) throws Exception {
                            int rnd = new Random().nextInt(channels.size());
                            NetworkServiceGrpc.NetworkServiceFutureStub stub = NetworkServiceGrpc
                                    .newFutureStub(channels.get(rnd));

                            ArrayList<Request> request = new ArrayList<>();
                            request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_READ,
                                    0, 0, key, 0,
                                    MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ""));
                            Packet requestPacket = RequestBuilder.buildPacket(request);
                            ListenableFuture<Packet> readListener = stub.readMetadata(requestPacket);
                            Packet packet = readListener.get();
                            Response response = packet.getResponse(0);
                            if (response.getStatus() == MessageConstants.STATUS_OK) {
                                Ack ack = response.getAck();
                                return new StorageModel(ack.getVersion(), ack.getDataSize(), ack.getKey(),
                                        MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ack.getCrc(), ack.getValue());
                            } else {
                                throw new IOException("Resource not found");
                            }
                        }
                    });

    }

    public LoadingCache<String, StorageModel> getCache() {
        return cache;
    }

    public void retrieveFollowerList() {
        NetworkServiceGrpc.NetworkServiceBlockingStub stub = NetworkServiceGrpc.newBlockingStub(this.masterChannel);
        ResponseList responseList = stub.getReadReplicaList(null);
        this.replicaSet = responseList.getFollowerList();
    }

    public void cleanup() {
        for (ManagedChannel channel: this.channels)
            channel.shutdown();
    }

    public void initStorage() {
        masterChannel = ManagedChannelBuilder.
                forAddress(ConfigConstants.HOST, ConfigConstants.PORT)
                .usePlaintext()
                .build();
        channels = new ArrayList<>();
        channels.add(masterChannel);

        retrieveFollowerList();

        for (Follower follower: replicaSet) {
            ManagedChannel channel = ManagedChannelBuilder.
                    forAddress(follower.getIp(), follower.getPort())
                    .usePlaintext()
                    .build();
            channels.add(channel);
        }
    }

    public ListenableFuture<Packet> create(String key, String value, int accessType) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_CREATE,
                0, value.length(), key,
                0, accessType, this.clientID, value));
        Packet packet = RequestBuilder.buildPacket(request);
        NetworkServiceGrpc.NetworkServiceFutureStub masterStub = NetworkServiceGrpc
                .newFutureStub(this.masterChannel);
        return masterStub.createMetadata(packet);
    }

    public ListenableFuture<StorageModel> read(String key) throws IOException {
        return new ListenableFuture<>() {
            StorageModel data = null;
            boolean isDone = false;
            @Override
            public void addListener(@Nonnull Runnable listener, @Nonnull Executor executor) {
                executor.execute(() -> {
                    try {
                        if (cacheEnabled) {
                            data = cache.get(key);
                        } else {
                            int rnd = new Random().nextInt(channels.size());
                            NetworkServiceGrpc.NetworkServiceFutureStub stub = NetworkServiceGrpc
                                    .newFutureStub(channels.get(rnd));

                            ArrayList<Request> request = new ArrayList<>();
                            request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_READ,
                                    0, 0, key, 0,
                                    MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ""));
                            Packet requestPacket = RequestBuilder.buildPacket(request);
                            ListenableFuture<Packet> readListener = stub.readMetadata(requestPacket);
                            Packet packet = readListener.get();
                            Response response = packet.getResponse(0);
                            if (response.getStatus() == MessageConstants.STATUS_OK) {
                                Ack ack = response.getAck();
                                data = new StorageModel(ack.getVersion(), ack.getDataSize(), ack.getKey(),
                                        MessageConstants.METADATA_ACCESS_PRIVATE, clientID,
                                        ack.getCrc(), ack.getValue());
                            } else {
                                throw new IOException("Resource not found");
                            }
                        }
                    } catch (Exception e) {
                        data = null;
                    }
                    isDone = true;
                    listener.run();
                });
            }

            @Override
            public boolean cancel(boolean b) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return false;
            }

            @Override
            public boolean isDone() {
                return isDone;
            }

            @Override
            public StorageModel get() throws InterruptedException, ExecutionException {
                return data;
            }

            @Override
            public StorageModel get(long l, @Nonnull TimeUnit timeUnit) throws InterruptedException, ExecutionException, TimeoutException {
                return data;
            }
        };
    }

    public ListenableFuture<Packet> update(String key, String value, int version) {
        return new ListenableFuture<>() {
            private Packet responsePacket = null;
            private boolean isDone = false;
            @Override
            public void addListener(@Nonnull Runnable listener, @Nonnull Executor executor) {
                executor.execute(() -> {
                    ArrayList<Request> request = new ArrayList<>();
                    request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_UPDATE,
                            version, value.length(), key,
                            0, MessageConstants.METADATA_ACCESS_PRIVATE, clientID, value));
                    Packet requestPacket = RequestBuilder.buildPacket(request);
                    NetworkServiceGrpc.NetworkServiceFutureStub masterStub = NetworkServiceGrpc
                            .newFutureStub(masterChannel);
                    ListenableFuture<Packet> res = masterStub.updateMetadata(requestPacket);
                    try {
                        this.responsePacket = res.get();
                        if (cacheEnabled && this.responsePacket
                                .getResponse(0).getStatus() == MessageConstants.STATUS_OK)
                            cache.invalidate(key);
                        this.isDone = true;
                        listener.run();
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }
                });
            }

            @Override
            public boolean cancel(boolean b) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return false;
            }

            @Override
            public boolean isDone() {
                return isDone;
            }

            @Override
            public Packet get() throws InterruptedException, ExecutionException {
                return this.responsePacket;
            }

            @Override
            public Packet get(long l, @Nonnull TimeUnit timeUnit) throws InterruptedException, ExecutionException, TimeoutException {
                return this.responsePacket;
            }
        };
    }

    public ListenableFuture<Packet> delete(String key, int version) {
        return new ListenableFuture<>() {
            private Packet responsePacket = null;
            private boolean isDone = false;
            @Override
            public void addListener(@Nonnull Runnable listener, @Nonnull Executor executor) {
                executor.execute(() -> {
                    ArrayList<Request> request = new ArrayList<>();
                    request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_DELETE,
                            version, 0, key,
                            0, MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ""));
                    Packet requestPacket = RequestBuilder.buildPacket(request);
                    NetworkServiceGrpc.NetworkServiceFutureStub masterStub = NetworkServiceGrpc
                            .newFutureStub(masterChannel);
                    ListenableFuture<Packet> res = masterStub.deleteMetadata(requestPacket);
                    try {
                        this.responsePacket = res.get();
                        if (cacheEnabled && responsePacket
                                .getResponse(0).getStatus() == MessageConstants.STATUS_OK)
                            cache.invalidate(key);
                        this.isDone = true;
                        listener.run();
                    } catch (InterruptedException | ExecutionException e) {
                        e.printStackTrace();
                    }
                });
            }

            @Override
            public boolean cancel(boolean b) {
                return false;
            }

            @Override
            public boolean isCancelled() {
                return false;
            }

            @Override
            public boolean isDone() {
                return this.isDone;
            }

            @Override
            public Packet get() throws InterruptedException, ExecutionException {
                return this.responsePacket;
            }

            @Override
            public Packet get(long l, @Nonnull TimeUnit timeUnit) throws InterruptedException, ExecutionException, TimeoutException {
                return this.responsePacket;
            }
        };
    }


}
