/*
 * Copyright 2021 Nilanjan Daw, Synerg Lab, Department of CSE, IIT Bombay
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package HpdosClient.lib;

import HpdosClient.ConfigConstants;
import HpdosClient.MessageFormat.MessageConstants;
import HpdosClient.MessageFormat.RequestBuilder;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import hpdos.grpc.*;
import io.grpc.ManagedChannel;
import io.grpc.ManagedChannelBuilder;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ExecutionException;

public class StorageServiceSync {

    private final String clientID;
    private ManagedChannel masterChannel;
    private ArrayList<ManagedChannel> channels;
    private List<Follower> replicaSet;
    private boolean cacheEnabled;
    private LoadingCache<String, StorageModel> cache = null;

    public StorageServiceSync(String clientID, boolean cacheEnabled) {
        this.clientID = clientID;
        this.cacheEnabled = cacheEnabled;
        if (this.cacheEnabled)
            this.cache = CacheBuilder
                    .newBuilder()
                    .maximumSize(10000)
                    .recordStats()
                    .build(new CacheLoader<>() {
                        @Override
                        public StorageModel load(@Nonnull String key) throws Exception {
                            int index = (int) (Math.random() * channels.size());
                            NetworkServiceGrpc.NetworkServiceBlockingStub stub =
                                    NetworkServiceGrpc.newBlockingStub(channels.get(index));

                            ArrayList<Request> request = new ArrayList<>();
                            request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_READ,
                                    0, 0, key, 0, MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ""));
                            Packet requestPacket = RequestBuilder.buildPacket(request);
                            Packet responsePacket = stub.readMetadata(requestPacket);
                            Response response = responsePacket.getResponse(0);
                            if (response.getStatus() == MessageConstants.STATUS_OK) {
                                Ack ack = response.getAck();
                                return new StorageModel(ack.getVersion(), ack.getDataSize(), ack.getKey(),
                                        MessageConstants.METADATA_ACCESS_PRIVATE, clientID, ack.getCrc(), ack.getValue());
                            } else {
                                throw new IOException("Resource not found");
                            }
                        }
                    });

    }

    public LoadingCache<String, StorageModel> getCache() {
        return cache;
    }

    public void retrieveFollowerList() {
        NetworkServiceGrpc.NetworkServiceBlockingStub stub = NetworkServiceGrpc.newBlockingStub(this.masterChannel);
        ResponseList responseList = stub.getReadReplicaList(null);
        this.replicaSet = responseList.getFollowerList();
    }

    public void cleanup() {
        for (ManagedChannel channel: this.channels)
            channel.shutdown();
    }

    public void initStorage() {
        masterChannel = ManagedChannelBuilder.
                forAddress(ConfigConstants.HOST, ConfigConstants.PORT)
                .usePlaintext()
                .build();
        channels = new ArrayList<>();
        channels.add(masterChannel);
        retrieveFollowerList();
        for (Follower follower: replicaSet) {
            ManagedChannel channel = ManagedChannelBuilder.
                    forAddress(follower.getIp(), follower.getPort())
                    .usePlaintext()
                    .build();
            channels.add(channel);
        }
    }

    // create a metadata block
    public long create(String key, String value) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_CREATE,
                0, value.length(), key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, value));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        NetworkServiceGrpc.NetworkServiceBlockingStub masterStub = NetworkServiceGrpc.newBlockingStub(masterChannel);
        Packet response = masterStub.createMetadata(packet);
        return timestampCreateStart;
    }

    // read back the metadata
    public Map.Entry<StorageModel, Long> read(String key) {
        if (cacheEnabled) {
            long timestampReadStart = System.currentTimeMillis();
            try {
                StorageModel model = cache.get(key);
                return new AbstractMap.SimpleEntry<>(model, timestampReadStart);
            } catch (ExecutionException e) {
                e.printStackTrace();
            }
            return new AbstractMap.SimpleEntry<>(null, timestampReadStart);
        } else {
            int rnd = new Random().nextInt(this.channels.size());
            NetworkServiceGrpc.NetworkServiceBlockingStub stub = NetworkServiceGrpc.newBlockingStub(channels.get(rnd));
            ArrayList<Request> request = new ArrayList<>();
            request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_READ,
                    0, 0, key, 0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, ""));
            Packet packet = RequestBuilder.buildPacket(request);
            long timestampReadStart = System.currentTimeMillis();
            Packet responsePacket = stub.readMetadata(packet);
            if (responsePacket.getResponse(0).getStatus() == MessageConstants.STATUS_OK) {
                StorageModel model = new StorageModel(responsePacket.getResponse(0).getAck(),
                        this.clientID, MessageConstants.METADATA_ACCESS_PRIVATE);
                return new AbstractMap.SimpleEntry<>(model, timestampReadStart);
            }
            return new AbstractMap.SimpleEntry<>(null, timestampReadStart);
        }
    }

    public long update(String key, String value, int version) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_UPDATE,
                version, value.length(), key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, value));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        NetworkServiceGrpc.NetworkServiceBlockingStub masterStub = NetworkServiceGrpc.newBlockingStub(masterChannel);
        Packet responsePacket = masterStub.updateMetadata(packet);
        if (this.cacheEnabled &&
                responsePacket.getResponse(0).getStatus() == MessageConstants.STATUS_OK)
            cache.invalidate(key);
        return timestampCreateStart;
    }

    public long delete(String key, int version) {
        ArrayList<Request> request = new ArrayList<>();
        request.add(RequestBuilder.buildRequest(MessageConstants.METADATA_DELETE,
                version, 0, key,
                0, MessageConstants.METADATA_ACCESS_PRIVATE, this.clientID, ""));
        Packet packet = RequestBuilder.buildPacket(request);
        long timestampCreateStart = System.currentTimeMillis();
        NetworkServiceGrpc.NetworkServiceBlockingStub masterStub = NetworkServiceGrpc.newBlockingStub(masterChannel);
        Packet responsePacket = masterStub.deleteMetadata(packet);
        if (this.cacheEnabled &&
                responsePacket.getResponse(0).getStatus() == MessageConstants.STATUS_OK)
            cache.invalidate(key);
        return timestampCreateStart;
    }
}
